@@ -3,7 +3,7 @@
   free available library PL/Vision. Please look www.quest.com
 
   Original author: Steven Feuerstein, 1996 - 2002
-  PostgreSQL implementation author: Pavel Stehule, 2006-2022
+  PostgreSQL implementation author: Pavel Stehule, 2006-2023
 
   This module is under BSD Licence
 
@@ -120,7 +120,6 @@ filterList(List *list, bool skip_spaces, bool qnames)
 {
 	List *result = NIL;
 	ListCell *cell;
-	bool isdot = false;
 	orafce_lexnode *a = NULL;
 	orafce_lexnode *dot = NULL;
 
@@ -130,7 +129,7 @@ filterList(List *list, bool skip_spaces, bool qnames)
 
 		if (qnames)
 		{
-			isdot = (IsType(nd, OTHERS) && (nd->str[0] == '.'));
+			bool		isdot = (IsType(nd, OTHERS) && (nd->str[0] == '.'));
 
 			if (IsType(nd, IDENT) && dot && a)
 			{
